/******************************************************************************
*
* panny.c - implementation of microstamp11 panny serial control (Panasonic TH50PH-D3)
*           for "Panasonic Control Program"
*
* Copyright (c) 2002, 2004, Don Woodward.
*
*
* This file is part of "Panasonic Control Program".
*
* "Panasonic Control Program" is free software; you can redistribute
* it and/or modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2 of the
* License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/

#include "microstamp11.h"
#include "panny.h"
#include "SonyIR.h"

/******************************************************************************
* panny.c private #defines
******************************************************************************/
#define kINITALCHARRESPONSETIMEOUT      (100)   // ~ 410 ms
#define kSUBSEQUENTCHARRESPONSETIMEOUT  (50)    // ~ 205 ms

/******************************************************************************
* panny.c private typedefs
******************************************************************************/
typedef struct _PannyCommand
{
    const char * m_pCommand;
    const char * m_pResponse;
} PannyCommand;

typedef struct _PannyMacroEntry
{
   PannyCommandID m_commandID;
   u8             m_nDelayMS;
} PannyMacroEntry;

/******************************************************************************
* panny.c private static variables
******************************************************************************/
static const PannyCommand s_commandTable[] = 
{
    { "PON",        "PON" },        // kCmdPanny_POWER_ON
    { "POF",        "POF" },        // kCmdPanny_POWER_OFF
    { "AUU",        "AUU" },        // kCmdPanny_VOL_UP
    { "AUD",        "AUD" },        // kCmdPanny_VOL_DOWN
    { "AMT",        "AMT" },        // kCmdPanny_VOL_MUTE
    { "IIS:VID",    "IIS" },        // kCmdPanny_INPUT_VIDEO
    { "IIS:YP1",    "IIS" },        // kCmdPanny_INPUT_COMPONENT
    { "IIS:RG1",    "IIS" },        // kCmdPanny_INPUT_VGA_PC
    { "DAM:NORM",   "DAM" },        // kCmdPanny_ASPECT_4x3
    { "DAM:ZOOM",   "DAM" },        // kCmdPanny_ASPECT_ZOOM
    { "DAM:FULL",   "DAM" },        // kCmdPanny_ASPECT_FULL
    { "DAM:JUST",   "DAM" },        // kCmdPanny_ASPECT_JUST
    { "DAM:SELF",   "DAM" },        // kCmdPanny_ASPECT_AUTO
};

static const PannyMacroEntry s_macroEntryOnVGAFull[] =          { {kCmdPanny_POWER_ON, 0}, {kCmdPanny_INPUT_VGA_PC, 0},     {kCmdPanny_ASPECT_FULL, 0}, { -1, 0 } };
static const PannyMacroEntry s_macroEntryOnComponentFull[] =    { {kCmdPanny_POWER_ON, 0}, {kCmdPanny_INPUT_COMPONENT, 0},  {kCmdPanny_ASPECT_FULL, 0}, { -1, 0 } };
static const PannyMacroEntry s_macroEntryOnVideoAuto[] =        { {kCmdPanny_POWER_ON, 0}, {kCmdPanny_INPUT_VIDEO, 0},      {kCmdPanny_ASPECT_AUTO, 0}, { -1, 0 } };
static const PannyMacroEntry s_macroEntryOff[] =                { {kCmdPanny_POWER_OFF, 0}, {-1, 0 } };

static const PannyMacroEntry * const s_macroTable[] =
{
    s_macroEntryOnVGAFull,          // kMacroPanny_ON_VGA_FULL
    s_macroEntryOnComponentFull,    // kMacroPanny_ON_COMPONENT_FULL
    s_macroEntryOnVideoAuto,        // kMacroPanny_ON_VIDEO_AUTO
    s_macroEntryOff,                // kMacroPanny_OFF
};

typedef enum _PannyDispatchType
{
    kDispatchTypeNull = 0,
    kDispatchTypeCommand = 1,
    kDispatchTypeMacro = 2,
} PannyDispatchType;

typedef struct _PannyIRDispatchEntry
{
    SonyIRDevice        m_nDevice;
    SonyIRKey           m_nKey;
    PannyDispatchType   m_dispatchType;
    union                                { PannyCommandID m_nCommandID; PannyMacroID m_nMacroID; }
                        m_commandOrMacro;              
} PannyIRDispatchEntry;

static const PannyIRDispatchEntry s_irDispatchTable[] =
{
    { SONY_IRD_LASERDISC,   SONY_IRK_1,             kDispatchTypeCommand,   { kCmdPanny_POWER_ON }              },
    { SONY_IRD_LASERDISC,   SONY_IRK_2,             kDispatchTypeCommand,   { kCmdPanny_POWER_OFF }             },
    { SONY_IRD_LASERDISC,   SONY_IRK_3,             kDispatchTypeCommand,   { kCmdPanny_INPUT_COMPONENT }       },
    { SONY_IRD_LASERDISC,   SONY_IRK_4,             kDispatchTypeCommand,   { kCmdPanny_INPUT_VIDEO }           },
    { SONY_IRD_LASERDISC,   SONY_IRK_5,             kDispatchTypeCommand,   { kCmdPanny_INPUT_VGA_PC }          },
    { SONY_IRD_LASERDISC,   SONY_IRK_6,             kDispatchTypeCommand,   { kCmdPanny_ASPECT_FULL }           },
    { SONY_IRD_LASERDISC,   SONY_IRK_7,             kDispatchTypeCommand,   { kCmdPanny_ASPECT_AUTO }           },
    { SONY_IRD_LASERDISC,   SONY_IRK_8,             kDispatchTypeCommand,   { kCmdPanny_ASPECT_ZOOM }           },
    { SONY_IRD_LASERDISC,   SONY_IRK_9,             kDispatchTypeCommand,   { kCmdPanny_ASPECT_4x3 }            },
    { SONY_IRD_LASERDISC,   SONY_IRK_0,             kDispatchTypeCommand,   { kCmdPanny_ASPECT_JUST }           },
    { SONY_IRD_LASERDISC,   SONY_IRK_PLUS10,        kDispatchTypeMacro,     { kMacroPanny_OFF }                 },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDPAUSE,       kDispatchTypeMacro,     { kMacroPanny_ON_COMPONENT_FULL }   },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDSTOP,        kDispatchTypeMacro,     { kMacroPanny_ON_VIDEO_AUTO }       },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDPLAY,        kDispatchTypeMacro,     { kMacroPanny_ON_VGA_FULL }         },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDREWIND,      kDispatchTypeNull,      { 0 }                               },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDFASTFORWARD, kDispatchTypeNull,      { 0 }                               },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDPREV,        kDispatchTypeNull,      { 0 }                               },
    { SONY_IRD_LASERDISC,   SONY_IRK_LDNEXT,        kDispatchTypeNull,      { 0 }                               },
    { SONY_IRD_LASERDISC,   SONY_IRK_POWER,         kDispatchTypeNull,      { 0 }                               },
    { 0,                    0,                      kDispatchTypeNull,      { 0 }                               },
};

static u8 s_nTimeout = 0;

/******************************************************************************
* panny.c private function forward declarations
******************************************************************************/

/******************************************************************************
* panny.c private static helper functions
******************************************************************************/
static boolean
CollectResponse(char * pResponse)
{
    s_nTimeout = 0;
    while (! serialCharPending()) if (s_nTimeout > kINITALCHARRESPONSETIMEOUT) return FALSE;
    *pResponse = serialGetChar(); if (*pResponse != 2) return FALSE;
    
    s_nTimeout = 0;
    while (! serialCharPending()) if (s_nTimeout > kSUBSEQUENTCHARRESPONSETIMEOUT) return FALSE;
    *pResponse++ = serialGetChar();

    s_nTimeout = 0;
    while (! serialCharPending()) if (s_nTimeout > kSUBSEQUENTCHARRESPONSETIMEOUT) return FALSE;
    *pResponse++ = serialGetChar();
    
    s_nTimeout = 0;
    while (! serialCharPending()) if (s_nTimeout > kSUBSEQUENTCHARRESPONSETIMEOUT) return FALSE;
    *pResponse = serialGetChar();

    s_nTimeout = 0;
    while (! serialCharPending()) if (s_nTimeout > kSUBSEQUENTCHARRESPONSETIMEOUT) return FALSE;
    if (3 != serialGetChar()) return FALSE;

    return TRUE;
}

/******************************************************************************
* panny.c public functions
******************************************************************************/
void
PannyOnRealTimeInterrupt()
{
    s_nTimeout++;
}

u8 s_nCommandSuccess = 0;
u8 s_nCommandFailure = 0;
u8 s_nMacroSuccess = 0;
u8 s_nMacroFailure = 0;

boolean
DispatchPannyCommand(PannyCommandID commandID)
{
    /* dispatches a discrete command */
    char response[4];
    const PannyCommand * pCommand = &s_commandTable[commandID];

    serialFlushInput();

    serialPutChar(2);
    serialPutString(pCommand->m_pCommand);
    serialPutChar(3);

    if (CollectResponse(response))
    {
        if ((response[0] == pCommand->m_pResponse[0]) &&
            (response[1] == pCommand->m_pResponse[1]) &&
            (response[2] == pCommand->m_pResponse[2]))
            {
                s_nCommandSuccess++;
                return TRUE;
            }
    }

    s_nCommandFailure++;
    return FALSE;
}

boolean
ExecutePannyMacro(PannyMacroID macroID)
{
    /* executes a sequence of commands */
    const PannyMacroEntry * pEntry = s_macroTable[macroID];

    while (pEntry->m_commandID != -1)
    {
        if (DispatchPannyCommand(pEntry->m_commandID))
        {
            s_nCommandSuccess--;
        }
        else
        {
            s_nCommandFailure--;
            s_nMacroFailure++;
            return FALSE;
        }
        // if (! DispatchPannyCommand(pEntry->m_commandID)) return FALSE;
        if (pEntry->m_nDelayMS) sleep(pEntry->m_nDelayMS);
        pEntry++;
    }

    s_nMacroSuccess++;
    return TRUE;
}

void
PannyNotifyIRCodeReceived(u16 nIRCode)
{
    const PannyIRDispatchEntry * pEntry = s_irDispatchTable;
    while (pEntry->m_dispatchType != kDispatchTypeNull)
    {
        if ((pEntry->m_nDevice == SONYIRCODETODEVICE(nIRCode)) && (pEntry->m_nKey == SONYIRCODETOKEY(nIRCode))) break;
        pEntry++;
    }

    switch (pEntry->m_dispatchType)
    {
        case kDispatchTypeCommand:
            DispatchPannyCommand(pEntry->m_commandOrMacro.m_nCommandID);
            break;
        case kDispatchTypeMacro:
            ExecutePannyMacro(pEntry->m_commandOrMacro.m_nMacroID);
            break;
        default:
            break;
    }
}

/******************************************************************************
*   LOG
*******************************************************************************
*   27-Jan-02   dwoodward   created
*
*/
