/******************************************************************************
*
* sonyIR.c - implementation of microstamp11 sonyIR decode for "Panasonic Control Program"
*
* Copyright (c) 2002, 2004, Don Woodward.
*
*
* This file is part of "Panasonic Control Program".
*
* "Panasonic Control Program" is free software; you can redistribute
* it and/or modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2 of the
* License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/

#include "microstamp11.h"
#include "hc11ed0.h"
#include "sonyIR.h"
#include "panny.h"

/******************************************************************************
* sonyIR.c private typedefs
******************************************************************************/
#define SIGNAL_IS_INVERTED_IE_5V_IS_LOW

#if defined(SIGNAL_IS_INVERTED_IE_5V_IS_LOW)
#define RISING_EDGE     EDG1B
#define FALLING_EDGE    EDG1A
#else
#define RISING_EDGE     EDG1A
#define FALLING_EDGE    EDG1B
#endif

#define kWAITING_FOR_HEADER_SILENCE         (20)
#define kWAITING_FOR_HEADER_START           (21)
#define kWAITING_FOR_HEADER_END             (22)
#define kDECODE_BIT_1                       (1)
#define kDECODE_BIT_12                      (12)
#define kWAITING_FOR_DELAY_AFTER_DECODE     (30)
#define kWAITING_FOR_IDLE_DISPATCH          (31)
#define kWAITING_FOR_DELAY_AFTER_DISPATCH   (32)

#define kSILENCE_BEFORE_HEADER              (6)     // require ~25ms of silence (in RTI ticks) before looking for a header
#define kSILENCE_AFTER_DECODE               (6)     // require ~25ms ms of silence (in RTI ticks) after the last decoded bit before dispatching the code
#define kSILENCE_DURING_DECODE_ERROR        (2)     // if 8ms of silence (in RTI ticks) is detected during any decode phase, error out
#define kSILENCE_AFTER_DECODE_DISPATCH      (25)    // require ~100 ms of silence (in RTI ticks) after a decode dispatch before looking for the next code
#define kRESET_REPEAT_TIMEOUT               (25)    // require an additional ~100ms break between the same code so as not to dispatch repeats

#define kTIC1_TICKS_PER_TIME_T              (1200)  // number of TIC1 ticks in 600us
#define kTIC1_TICKS_PER_2T_LO               (2064)  // swing 14% each way!
#define kTIC1_TICKS_PER_2T_HI               (2736)
#define kTIC1_TICKS_PER_3T_LO               (3096)  // swing 14% each way!
#define kTIC1_TICKS_PER_3T_HI               (4104)
#define kTIC1_TICKS_PER_4T_LO               (4128)  // swing 14% each way!
#define kTIC1_TICKS_PER_4T_HI               (5472)

/* Each TIC1 tick is 500 nanoseconds or .5 microseconds */
//    1TIC = .5 us
//       T = 600us
//      4T = 2400us
// 1200TIC = 1T
// 4800TIC = 4T

/******************************************************************************
* sonyIR.c private static variables
******************************************************************************/
static u8  s_state = kWAITING_FOR_HEADER_SILENCE;
static u8  s_nSilenceTicks = 0;
static u16 s_nTimeLastEdge = 0;
static u16 s_nIRCode = 0;
static u16 s_nLastIRCode = 0;

/******************************************************************************
* sonyIR.c debug functionality
******************************************************************************/
//#define _DEBUG_IR
#ifdef _DEBUG_IR
extern void NumberToStringHex(u16 nNumber, char * pBuff, u8 nBuffLen);

typedef struct _KeyNameEntry
{
    const SonyIRDevice m_nKey;
    const char * m_pKeyName;
} KeyNameEntry;

typedef struct _DevNameEntry
{
    const SonyIRDevice m_nDevice;
    const char * m_pDevName;
    const KeyNameEntry * m_pKeyNames;
} DevNameEntry;

static const KeyNameEntry s_keyNamesCommon[] =
{
    { SONY_IRK_POWER,               "Power" },
    { SONY_IRK_1,                   "1" },
    { SONY_IRK_2,                   "2" },
    { SONY_IRK_3,                   "3" },
    { SONY_IRK_4,                   "4" },
    { SONY_IRK_5,                   "5" },
    { SONY_IRK_6,                   "6" },
    { SONY_IRK_7,                   "7" },
    { SONY_IRK_8,                   "8" },
    { SONY_IRK_9,                   "9" },
    { SONY_IRK_0,                   "0" },
    { SONY_IRK_ENTER,               "Enter" },
    { SONY_IRK_CHANNELUP,           "Channel Up" },
    { SONY_IRK_CHANNELDOWN,         "Channel Down" },
    { 0,                            0 },
};

static const KeyNameEntry s_keyNamesTV[] =
{
    { SONY_IRK_TVMUTE,              "Mute" },
    { SONY_IRK_TVSLEEP,             "Sleep" },
    { SONY_IRK_TVCLEAR,             "Clear" },
    { SONY_IRK_TVTIMERBLOCK,        "Timer Block" },
    { SONY_IRK_TVRESET,             "Reset" },
    { SONY_IRK_TVSELECT,            "Select" },
    { SONY_IRK_TVOFFREPEAT,         "Off Repeat" },
    { SONY_IRK_TVAMPM,              "AM/PM" },
    { SONY_IRK_TVLEVELDOWN,         "Level Down" },
    { SONY_IRK_TVLEVELUP,           "Level Up" },
    { SONY_IRK_TVERASE,             "Erase" },
    { SONY_IRK_TVADD,               "Add" },
    { SONY_IRK_TVAUTOPROGRAM,       "Auto Program" },
    { SONY_IRK_TVCABLE,             "Cable" },
    { SONY_IRK_TVTVVIDEO,           "TV/Video" },
    { SONY_IRK_TVMTS,               "MTS" },
    { SONY_IRK_TVDISPLAY,           "Display" },
    { SONY_IRK_TVJUMP,              "Jump" },
    { SONY_IRK_TVPICTUREUP,         "Picture Up" },
    { SONY_IRK_TVPICTUREDOWN,       "Picture Down" },
    { SONY_IRK_TVVOLUMEUP,          "Volume Up" },
    { SONY_IRK_TVVOLUMEDOWN,        "Volume Down" },
    { 0,                            0 },
};

static const KeyNameEntry s_keyNamesCD[] =
{
    { SONY_IRK_CDCONTINUE,          "Continue" },
    { SONY_IRK_CDSHUFFLE,           "Shuffle" },
    { SONY_IRK_CDPROGRAM,           "Program" },
    { SONY_IRK_CDMEGACONTROL,       "Mega Control" },
    { SONY_IRK_CDMUSICSCAN,         "Music Scan" },
    { SONY_IRK_CDHIGHLIGHT,         "Highlight" },
    { SONY_IRK_CDDISC,              "Disc" },
    { SONY_IRK_CDTIMETEXT,          "Time/Text" },
    { SONY_IRK_CDFADER,             "Fader" },
    { SONY_IRK_CDGREATERTHAN10,     "> 10" },
    { SONY_IRK_CD10,                "10" },
    { SONY_IRK_CDREPEAT,            "Repeat" },
    { SONY_IRK_CDCHECK,             "Check" },
    { SONY_IRK_CDCLEAR,             "Clear" },
    { SONY_IRK_CDMEMO,              "Memo" },
    { SONY_IRK_CDPLAY,              "Play" },
    { SONY_IRK_CDPAUSE,             "Pause" },
    { SONY_IRK_CDSTOP,              "Stop" },
    { SONY_IRK_CDPREV,              "Prev" },
    { SONY_IRK_CDNEXT,              "Next" },
    { SONY_IRK_CDREW,               "Rewind" },
    { SONY_IRK_CDFF,                "Fast Forward" },
    { SONY_IRK_CDDISCMINUS,         "Disc Minus" },
    { SONY_IRK_CDDISCPLUS,          "Disc Plus" },
    { SONY_IRK_CDLINEOUTMINUS,      "LineOut Minus" },
    { SONY_IRK_CDLINEOUTPLUS,       "LineOut Plus" },
    { 0,                            0 },
};

static const KeyNameEntry s_keyNamesVCR[] =
{
    { SONY_IRK_VCRTIMERREC,         "Timer Record" },
    { SONY_IRK_VCRTVVCR,            "TV/VCR" },
    { SONY_IRK_VCREDITMONITOR,      "Edit Monitor" },
    { SONY_IRK_VCRTIMERONSCREEN,    "Timer OnScreen" },
    { SONY_IRK_VCRINDEX,            "Index" },
    { SONY_IRK_VCRDATASCREEN,       "Data Screen" },
    { SONY_IRK_VCRAUDIOMONITOR,     "Audio Monitor" },
    { SONY_IRK_VCRCOUNTERRESET,     "Counter Reset" },
    { SONY_IRK_VCRSHUTTLEEDITBACK,  "Shuttle Edit Back" },
    { SONY_IRK_VCRSHUTTLEEDITFWD,   "Shuttle Edit Forward" },
    { SONY_IRK_VCRPAUSE,            "Pause" },
    { SONY_IRK_VCRRECORD,           "Record" },
    { SONY_IRK_VCRAUDIOINSERT,      "Audio Insert" },
    { SONY_IRK_VCRVIDEOINSERT,      "Video Insert" },
    { SONY_IRK_VCRHIGHSPEEDREW,     "High Speed Rewind" },
    { SONY_IRK_VCRSEARCHBACK,       "Search Back" },
    { SONY_IRK_VCRSEARCHFORWARD,    "Search Forward" },
    { SONY_IRK_VCRREWIND,           "Rewind" },
    { SONY_IRK_VCRFASTFORWARD,      "Fast Forward" },
    { SONY_IRK_VCRPLAY,             "Play" },
    { SONY_IRK_VCRSTOP,             "Stop" },
    { SONY_IRK_VCRMENU,             "Menu" },
    { SONY_IRK_VCRNORTH,            "North" },
    { SONY_IRK_VCRSOUTH,            "South" },
    { SONY_IRK_VCRWEST,             "West" },
    { SONY_IRK_VCREAST,             "East" },
    { SONY_IRK_VCREXECUTE,          "Execute" },
    { SONY_IRK_VCRFUNCTIONMEMORY,   "Function Memory" },
    { SONY_IRK_VCRTVSCAN,           "TV Scan" },
    { SONY_IRK_VCRPIP,              "PIP" },
    { SONY_IRK_VCRSHIFT,            "Shift" },
    { SONY_IRK_VCRDIGITALOFF,       "Digital Off" },
    { SONY_IRK_VCRFRAMEBACK,        "Frame Back" },
    { SONY_IRK_VCRFRAMESTILL,       "Frame Still" },
    { SONY_IRK_VCRFRAMEFORWARD,     "Frame Forward" },
    { SONY_IRK_VCRTIMERCLEAR,       "Timer Clear" },
    { SONY_IRK_VCRRECORDMODE,       "Record Mode" },
    { SONY_IRK_VCRINPUTSELECT,      "Input Select" },
    { SONY_IRK_VCRSLOWMINUS,        "Slow Minus" },
    { SONY_IRK_VCRSLOWPLUS,         "Slow Plus" },
    { SONY_IRK_VCR2XPLAYBACK,       "Play x2" },
    { SONY_IRK_VCRTIMERCHECK,       "Timer Check" },
    { SONY_IRK_VCRINDEXMARK,        "Index Mark" },
    { SONY_IRK_VCRINDEXERASE,       "Index Erase" },
    { 0,                            0 },
};

static const KeyNameEntry s_keyNamesLASERDISC[] =
{
    { SONY_IRK_LDREWIND,            "Rewind" },
    { SONY_IRK_LDFASTFORWARD,       "Fast Forward" },
    { SONY_IRK_LDPREV,              "Previous" },
    { SONY_IRK_LDNEXT,              "Next" },
    { SONY_IRK_LDPAUSE,             "Pause" },
    { SONY_IRK_LDSTOP,              "Stop" },
    { SONY_IRK_LDPLAY,              "Play" },
    { SONY_IRK_PLUS10,              "+ 10" },
    { 0,                            0 },
};

static const DevNameEntry s_devNames[] =
{
    { SONY_IRD_CD,          "CD Player",    s_keyNamesCD },
    { SONY_IRD_TV,          "TV",           s_keyNamesTV },
    { SONY_IRD_VCR1,        "VCR1",         s_keyNamesVCR },
    { SONY_IRD_VCR2,        "VCR2",         s_keyNamesVCR },
    { SONY_IRD_VCR3,        "VCR3",         s_keyNamesVCR },
    { SONY_IRD_LASERDISC,   "Laser Disc",   s_keyNamesLASERDISC },
    { 0,                    0,              0 },
};

static const char * GetDeviceName()
{
    const DevNameEntry * pEntry = s_devNames;
    while (pEntry->m_pDevName)
    {
        if (pEntry->m_nDevice == SONYIRCODETODEVICE(s_nIRCode)) break;
        pEntry++;
    }
    return (pEntry->m_pDevName ? pEntry->m_pDevName : "Unknown");
}

static const char * GetKeyName()
{
    const char * pKeyName = NULL;
    const KeyNameEntry * pKeyNameEntry = NULL;
    const DevNameEntry * pEntry = s_devNames;
    while (pEntry->m_pDevName)
    {
        if (pEntry->m_nDevice == SONYIRCODETODEVICE(s_nIRCode))
        {
            pKeyNameEntry = pEntry->m_pKeyNames;
            break;
        }
        pEntry++;
    }

again:
    while (pKeyNameEntry->m_pKeyName)
    {
        if (pKeyNameEntry->m_nKey == SONYIRCODETOKEY(s_nIRCode))
        {
            pKeyName = pKeyNameEntry->m_pKeyName;
            break;
        }
        pKeyNameEntry++;
    }

    if (pKeyName == NULL)
    {
        pKeyName = "Unknown"; // so we only goto again one time
        pKeyNameEntry = s_keyNamesCommon;
        goto again;
    }

    return pKeyName;
}

static void
DumpIRCode()
{
//  [Code/Dev/Key = A10/10/50] [Name = TV:6]
    char buff[6];
    NumberToStringHex(s_nIRCode, buff, sizeof(buff));
    serialPutString("[Code/Dev/Key = "); serialPutString(buff); serialPutChar('/');
    NumberToStringHex(SONYIRCODETODEVICE(s_nIRCode), buff, sizeof(buff));
    serialPutString(buff); serialPutChar('/');
    NumberToStringHex(SONYIRCODETOKEY(s_nIRCode), buff, sizeof(buff));
    serialPutString(buff);
    serialPutString("] [Name = ");
    serialPutString(GetDeviceName()); serialPutChar(':');
    serialPutString(GetKeyName());
    serialPutString("]\n");
}

#endif // #ifdef _DEBUG_IR

/******************************************************************************
* sonyIR.c private static helper functions
******************************************************************************/
static void
TriggerOnRisingEdge()
{
    u8 reg;
    TFLG1 = IC1F;           // clear the T      IC1 flag

    // configure TIC1 for rising edge interrupts
    reg = TCTL2;
    reg &= ~(RISING_EDGE | FALLING_EDGE);
    reg |= RISING_EDGE;
    TCTL2 = reg;

    // request TIC1 interrupts
    reg = TMSK1;
    reg |= IC1I;
    TMSK1 = reg;
}

static void
TriggerOnFallingEdge()
{
    u8 reg;
    TFLG1 = IC1F;           // clear the TIC1 flag

    // configure TIC1 for rising edge interrupts
    reg = TCTL2;
    reg &= ~(RISING_EDGE | FALLING_EDGE);
    reg |= FALLING_EDGE;
    TCTL2 = reg;

    // request TIC1 interrupts
    reg = TMSK1;
    reg |= IC1I;
    TMSK1 = reg;
}

static u16
ConvertElapsedTicksToTimeT(u16 nElapsedTicks)
{
         if (nElapsedTicks > kTIC1_TICKS_PER_4T_LO) return (nElapsedTicks < kTIC1_TICKS_PER_4T_HI) ? 4 : 0;
    else if (nElapsedTicks > kTIC1_TICKS_PER_3T_LO) return (nElapsedTicks < kTIC1_TICKS_PER_3T_HI) ? 3 : 0;
    else if (nElapsedTicks > kTIC1_TICKS_PER_2T_LO) return (nElapsedTicks < kTIC1_TICKS_PER_2T_HI) ? 2 : 0;
    else return 0;
}

/******************************************************************************
* sonyIR.c public functions
******************************************************************************/
void
SonyIR_Reset()
{
    s_state = kWAITING_FOR_HEADER_SILENCE;
    s_nSilenceTicks = 0;
    s_nTimeLastEdge = 0;
    s_nIRCode = 0;
    TriggerOnRisingEdge();
}

void
SonyIR_OnRealTimeInterrupt()
{
    s_nSilenceTicks++;
    switch (s_state)
    {
        case kWAITING_FOR_HEADER_SILENCE:
            if (s_nSilenceTicks >= kSILENCE_BEFORE_HEADER) s_state = kWAITING_FOR_HEADER_START;
            break;
        case kWAITING_FOR_HEADER_START:
            if (s_nSilenceTicks >= kRESET_REPEAT_TIMEOUT) s_nLastIRCode = 0;
            break;
        case kWAITING_FOR_HEADER_END:
            if (s_nSilenceTicks >= kSILENCE_DURING_DECODE_ERROR)
            {
                s_state = 98;
                SonyIR_Reset();
            }
            break;
        case kWAITING_FOR_DELAY_AFTER_DECODE:
            if (s_nSilenceTicks >= kSILENCE_AFTER_DECODE)
            {
                if (s_nLastIRCode != s_nIRCode)
                {
                    s_nLastIRCode = s_nIRCode;
                    s_state = kWAITING_FOR_IDLE_DISPATCH;
                }
                else
                {
                    s_state = kWAITING_FOR_DELAY_AFTER_DISPATCH;
                    s_nSilenceTicks = 0;
                    s_nTimeLastEdge = 0;
                    s_nIRCode = 0;
                }
            }
            break;
        case kWAITING_FOR_IDLE_DISPATCH:
            break;
        case kWAITING_FOR_DELAY_AFTER_DISPATCH:
            if (s_nSilenceTicks >= kSILENCE_AFTER_DECODE) SonyIR_Reset();
            break;
        default:
            if (s_nSilenceTicks >= kSILENCE_DURING_DECODE_ERROR) SonyIR_Reset();
            break;
    }
}

void
SonyIR_OnTimerInputCaptureInterrupt()
{
    u16 nTimeThisEdge = TIC1;
    u16 nTimeDiff;

    // reset the silence timer
    s_nSilenceTicks = 0;

    switch (s_state)
    {
        case kWAITING_FOR_HEADER_SILENCE:
            // we got a rising edge while waiting for silence before the header
            // continue waiting for silence (from scratch)
            return;
        case kWAITING_FOR_HEADER_START:
            // we got our rising edge for the header start,
            // remember when this occurred and trigger on falling edge for the end of the header
            s_nTimeLastEdge = nTimeThisEdge;
            s_state = kWAITING_FOR_HEADER_END;
            TriggerOnFallingEdge();
            return;
        case kWAITING_FOR_DELAY_AFTER_DECODE:
            // we got a rising edge after we finished decoding, but before the silence after period ended
            // ignore this decode
            SonyIR_Reset();
            return;
        case kWAITING_FOR_IDLE_DISPATCH:
        case kWAITING_FOR_DELAY_AFTER_DISPATCH:
            // we got a rising edge after we dispatched the code, but before the delay after the dispatch
            // continue waiting for the delay (from scratch)
            return;
        default:
            // handle the other cases below
            break;
    }

    // we got a falling edge, calculate the time from the last edge
    if (nTimeThisEdge > s_nTimeLastEdge)
    {
        nTimeDiff = nTimeThisEdge - s_nTimeLastEdge;
    }
    else
    {
        nTimeDiff = nTimeThisEdge + (0xFFFF - s_nTimeLastEdge);
    }

    // remember this edge as the last edge for next time
    s_nTimeLastEdge = nTimeThisEdge;

    // calculate the number of ticks of time T (600us) and put it innTimeThisEdge
    nTimeThisEdge = ConvertElapsedTicksToTimeT(nTimeDiff);

    switch (s_state)
    {
        case kWAITING_FOR_HEADER_END:
            // the header is 4T
            if (nTimeThisEdge == 4) s_state = kDECODE_BIT_1;
            else SonyIR_Reset();
            break;
        case kDECODE_BIT_12:
            // add in bit 12, reset TIC1 to trigger on rising edge, set state to waiting
            s_nIRCode <<= 1;
            s_state = kWAITING_FOR_DELAY_AFTER_DECODE;
            switch (nTimeThisEdge)
            {
                case 2:             // 0 bit
                    TriggerOnRisingEdge();
                    break;
                case 3:             // 1 bit
                    s_nIRCode++;
                    TriggerOnRisingEdge();
                    break;
                default:           // invalid
                    SonyIR_Reset();
                    break;            
            }
            break;
        default:
            // add in bit, increment state
            s_nIRCode <<= 1;
            s_state++;
            switch (nTimeThisEdge)
            {
                case 2:             // 0 bit
                    break;
                case 3:             // 1 bit
                    s_nIRCode++;
                    break;
                default:            // invalid
                    SonyIR_Reset();
                    break;            
            }
            break;
    }
}

void
SonyIR_OnIdle()
{
    disable();
    if (s_state == kWAITING_FOR_IDLE_DISPATCH)
    {
        enable();
        #ifdef _DEBUG_IR
            DumpIRCode();
        #else
            PannyNotifyIRCodeReceived(s_nIRCode);
        #endif
        disable();
        s_state = kWAITING_FOR_DELAY_AFTER_DISPATCH;
        s_nSilenceTicks = 0;
        s_nTimeLastEdge = 0;
        s_nIRCode = 0;
    }
    enable();
}

/******************************************************************************
*   LOG
*******************************************************************************
*   07-Jan-02   dwoodward   created
*
*/
